# Invoca::Metrics

Publish metrics for observability!

## Dependencies
* Ruby >= 2.6
* ActiveSupport >= 4.2, < 7

## Usage

Mix in the `Invoca::Metrics::Source` module:
```ruby
class MyClass
  include Invoca::Metrics::Source
  ...
end
```

Then declare the metrics that can be published by the class using a `declare_metrics` DSL block.
Each entry in the block starts with the metric type as the DSL method:

| Metric Type | Value                                               | Metric Instance Methods                                                                                                                                                                                                                                            |
|:-----------:|:----------------------------------------------------|:-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
|  `counter`  | monotonically incrementing floating point           | `.increment(value = 1, **labels)`                                                                                                                                                                                                                                  |
|   `gauge`   | arbitrary floating point                            | `.set(value, **labels)`                                                                                                                                                                                                                                            |
| `histogram` | histogram counters with configurable bucket cutoffs | `.add(value, **labels)`<br/><br/>`.time(return_timing: false, **labels, &block)`<br/>yields to block and `add()`s the milliseconds spent to the histogram; returns either `block_value` or `[milliseconds, block_value]` depending on the `return_timing` argument |

After the metric type, the first argument is the metric name, given as a symbol. This is globally unique.
Note that other classes in the process are allowed to declare the same metric, as long as the type and labels are the same.

Each metric may have any number of labels declared in the optional `labels:` hash, with default values in case that label is not passed when the metric is published.

Note: An implicit label of `service: <service_name>` is always added. The `<service_name>` is a configured value.

Histograms have a required argument, `buckets: <array>`, where the array is the histogram bucket cutoffs.

```ruby
class MyClass
  include Invoca::Metrics::Source

  declare_metrics do
    counter   :request_total, labels: { status: '200' }
    gauge     :available_memory
    histogram :processing_time, buckets: [30.seconds, 1.hour, 2.hours, 4.hours, 5.hours, 6.hours, 10.hours]
  end

  ...
end
```

After declaring the metrics, you may use `prometheus_metrics.<metric_name>` (in either the class or instance) to publish the metric.
```ruby
prometheus_metrics.request_total.increment       # increment the counter by 1

prometheus_metrics.request_total.increment(5)    # increment the counter by 5

prometheus_metrics.available_memory.set(10.25)   # set the gauge to 10.25

prometheus_metrics.processing_time.add(100)      # Increment the bucket closest to the value 100
prometheus_metrics.processing_time.time { work } # measure the elapsed milliseconds to run the given block and increment the bucket closest to that time period in milliseconds
```

## Installation

Add this line to your application's Gemfile:

    gem 'invoca-metrics'

And then execute:

    $ bundle

Or install it yourself as:

    $ gem install invoca-metrics

## Setup

### Default Setup

Add the following code to your application...

    require 'invoca/metrics'

    Invoca::Metrics.service_name    = "tts-service"
    Invoca::Metrics.server_name     = Socket.gethostname
    Invoca::Metrics.cluster_name    = "production"
    Invoca::Metrics.sub_server_name = "worker_process_1"
    Invoca::Metrics.statsd_host     = "255.0.0.123"
    Invoca::Metrics.statsd_port     = 200

    Invoca::Metrics::Client.logger = logger
    Invoca::Metrics::Client.log_send_failures = ['staging', 'production'].include(environment)

Note:  the `service_name` setting is required; all others are optional.

### Multi Configuration

To set multiple configurations, supply a `config` hash with additional settings.

    Invoca::Metrics.config = {
      deployment_group: {
        server_name: "deployment"
        statsd_host: "255.0.0.456",
        statsd_port: 300
      },
      region: {
        server_name:     "region_name"
        statsd_host:     "255.0.0.789",
        statsd_port:     400,
        sub_server_name: nil
        cluster_name:    nil
      }
    }

The settings (`[service_name, server_name, cluster_name, sub_server_name, statsd_host, statsd_port]`) directly set on `Invoca::Metrics` will be the default values supplied if the individual configuration does not supply the option.
It's highly suggested that each configuration has its own `statsd_host` and `statsd_port` along with unique naming since writing the same metric from one statsd node could be overwritten by the same metric from a separate node.

In order to set a configuration as the default, set the configuration key as `default_config_key`.

    Invoca::Metrics.default_config_key = :deployment_group

Any keys missing from the `default_config_key` config will by default have the values from the keys set directly on `Invoca::Metrics`.

The full set of default values for the above example would then be

    service_name:    "my_event_worker"
    cluster_name:    "production"
    sub_server_name: "worker_process_1"
    server_name:     "deployment"
    statsd_host:     "255.0.0.456",
    statsd_port:     300

### Prometheus Metric Exporting Setup
There are 2 ways to set up metrics for being scraped by Prometheus: `:direct` and `:push`).
`:direct` will expose a web `server directly from the running process in order for metrics to be scraped while `:push` will push metrics over to an aggregate gateway which in turn implements a web server.

#### `:direct` Configuration Example
```ruby
Invoca::Metrics::Prometheus.configure do |config|
  config.mode = :direct
  config.port = 9394 # Port to expose metrics from
end
```

#### `:push` Configuration
```ruby
Invoca::Metrics::Prometheus.configure do |config|
  config.mode = :push
  config.host = '127.0.0.1' # IP address to push metrics to
  config.port = 9394 # Port to push metrics to
end
```

## Contributing

1. Fork it
2. Create your feature branch (`git checkout -b my-new-feature`)
3. Commit your changes (`git commit -am 'Add some feature'`)
4. Push to the branch (`git push origin my-new-feature`)
5. Create new Pull Request
